diff --git a/sys/netinet/tcp_input.c b/sys/netinet/tcp_input.c
index 75609fd..0d11d9f 100644
--- a/sys/netinet/tcp_input.c
+++ b/sys/netinet/tcp_input.c
@@ -145,6 +145,12 @@ SYSCTL_VNET_INT(_net_inet_tcp, OID_AUTO, drop_synfin, CTLFLAG_RW,
     &VNET_NAME(drop_synfin), 0,
     "Drop TCP packets with SYN+FIN set");
 
+VNET_DEFINE(int, tcp_do_ncwv) = 0;
+#define V_tcp_do_ncwv  VNET(tcp_do_ncwv)
+SYSCTL_VNET_INT(_net_inet_tcp, OID_AUTO, do_ncwv, CTLFLAG_RW,
+    &VNET_NAME(tcp_do_ncwv), 0,
+    "Do New-CWV targeted to rate-limited applications");
+
 VNET_DEFINE(int, tcp_do_rfc3042) = 1;
 #define	V_tcp_do_rfc3042	VNET(tcp_do_rfc3042)
 SYSCTL_VNET_INT(_net_inet_tcp, OID_AUTO, rfc3042, CTLFLAG_RW,
@@ -228,6 +234,7 @@ static void	 tcp_dropwithreset(struct mbuf *, struct tcphdr *,
 static void	 tcp_pulloutofband(struct socket *,
 		     struct tcphdr *, struct mbuf *, int);
 static void	 tcp_xmit_timer(struct tcpcb *, int);
+static void  ncwv_check_phase(struct tcpcb *, struct tcphdr *);
 static void	 tcp_newreno_partial_ack(struct tcpcb *, struct tcphdr *);
 static void inline 	tcp_fields_to_host(struct tcphdr *);
 #ifdef TCP_SIGNATURE
@@ -289,6 +296,7 @@ static void inline
 cc_ack_received(struct tcpcb *tp, struct tcphdr *th, uint16_t type)
 {
 	INP_WLOCK_ASSERT(tp->t_inpcb);
+	int use_cc_algo=1;
 
 	tp->ccv->bytes_this_ack = BYTES_THIS_ACK(tp, th);
 	if (tp->snd_cwnd <= tp->snd_wnd)
@@ -310,7 +318,12 @@ cc_ack_received(struct tcpcb *tp, struct tcphdr *th, uint16_t type)
 		}
 	}
 
-	if (CC_ALGO(tp)->ack_received != NULL) {
+	if (V_tcp_do_ncwv) {
+		ncwv_check_phase(tp,th);
+		if (!IN_VALPHASE(tp->t_flags))
+			use_cc_algo = 0;
+	}
+	if (use_cc_algo && CC_ALGO(tp)->ack_received != NULL) {
 		/* XXXLAS: Find a way to live without this */
 		tp->ccv->curack = th->th_ack;
 		CC_ALGO(tp)->ack_received(tp->ccv, type);
@@ -384,6 +397,10 @@ cc_conn_init(struct tcpcb *tp)
 			tp->snd_cwnd = 4 * tp->t_maxseg;
 	}
 
+	if (V_tcp_do_ncwv) {
+		tp->max_ack_prev = tp->iss;
+		tp->IW = tp->snd_cwnd;
+	}
 	if (CC_ALGO(tp)->conn_init != NULL)
 		CC_ALGO(tp)->conn_init(tp->ccv);
 }
@@ -433,6 +450,28 @@ cc_cong_signal(struct tcpcb *tp, struct tcphdr *th, uint32_t type)
 		break;
 	}
 
+	/*
+	 *Exit the NVP, that means, stop the 5 min counter.
+	 *Flag on unset and stopped timer means that
+	 *we have exited the NVP. Set pipeAck to max value.
+	 *Record Flightsize so as to use it at the end
+	 *of the congestion.
+	 */
+	if (V_tcp_do_ncwv && type != CC_DUPACK && type != CC_RTO_ERR) {
+		/*
+		 *Exit the NVP, that means, stop the 5 min counter.
+		 *Flag on unset and stopped timer means that
+		 *we have exited the NVP. Set pipeAck to max value.
+		 *Record Flightsize so as to use it at the end
+		 *of the congestion.
+		 */
+		tp->lossflightsize = tp->snd_max - tp->snd_una;
+		tcp_timer_activate(tp, TT_NVP, 0);
+		EXIT_VALPHASE(tp->t_flags);
+		tp->pipeack = TCP_MAXWIN << TCP_MAX_WINSHIFT;
+		if(type != CC_RTO)
+			tp->snd_cwnd = min(tp->snd_cwnd/2,max(tp->pipeack,tp->lossflightsize));
+	}
 	if (CC_ALGO(tp)->cong_signal != NULL) {
 		if (th != NULL)
 			tp->ccv->curack = th->th_ack;
@@ -451,6 +490,16 @@ cc_post_recovery(struct tcpcb *tp, struct tcphdr *th)
 		tp->ccv->curack = th->th_ack;
 		CC_ALGO(tp)->post_recovery(tp->ccv);
 	}
+	if (V_tcp_do_ncwv)
+		/*
+		 * Fast recovery will conclude after returning from this
+		 * function. Reset the cwin to the value specified by the draft,
+		 * cwnd = ((FlightSize - R)/2), if SACK is used, standard behaviour
+		 * otherwise and also reset pipeACK
+		 */
+		tp->pipeack = TCP_MAXWIN << TCP_MAX_WINSHIFT;
+		if ((tp->t_flags & TF_SACK_PERMIT))
+			tp->snd_cwnd = (tp->lossflightsize - tp->sackhint.sack_bytes_rexmit) / 2;
 	/* XXXLAS: EXIT_RECOVERY ? */
 	tp->t_bytes_acked = 0;
 }
@@ -3692,6 +3741,30 @@ tcp_mssopt(struct in_conninfo *inc)
 	return (mss);
 }
 
+/*
+ * Check whether the sender lies in the Validated or
+ * non-Validate period.
+ */
+static void
+ncwv_check_phase(struct tcpcb *tp, struct tcphdr *th)
+{
+	INP_WLOCK_ASSERT(tp->t_inpcb);
+
+	if (th->th_ack >= tp->max_ack_prev) {
+		tp->pipeack = tp->snd_max - tp->snd_una;
+		tp->max_ack_prev = th->th_ack;
+		tcp_timer_activate(tp, TT_PACK, max(3*tp->t_srtt,1000));
+	}
+	if (tp->pipeack >= (tp->snd_cwnd / 2)){
+		ENTER_VALPHASE(tp->t_flags);
+		if(tcp_timer_active(tp, TT_NVP))
+			tcp_timer_activate(tp, TT_NVP, 0);
+	} else {
+		EXIT_VALPHASE(tp->t_flags);
+		if(!tcp_timer_active(tp, TT_NVP))
+			tcp_timer_activate(tp, TT_NVP, 300000);
+	}
+}
 
 /*
  * On a partial ack arrives, force the retransmission of the
diff --git a/sys/netinet/tcp_output.c b/sys/netinet/tcp_output.c
index 00d5415..fc37006 100644
--- a/sys/netinet/tcp_output.c
+++ b/sys/netinet/tcp_output.c
@@ -158,8 +158,10 @@ cc_after_idle(struct tcpcb *tp)
 {
 	INP_WLOCK_ASSERT(tp->t_inpcb);
 
-	if (CC_ALGO(tp)->after_idle != NULL)
+	if (!V_tcp_do_ncwv && CC_ALGO(tp)->after_idle != NULL)
 		CC_ALGO(tp)->after_idle(tp->ccv);
+	else if(V_tcp_do_ncwv)
+		tp->pipeack = 0;
 }
 
 /*
diff --git a/sys/netinet/tcp_subr.c b/sys/netinet/tcp_subr.c
index 5d37b50..bd4151e 100644
--- a/sys/netinet/tcp_subr.c
+++ b/sys/netinet/tcp_subr.c
@@ -783,6 +783,8 @@ tcp_newtcpcb(struct inpcb *inp)
 	callout_init(&tp->t_timers->tt_keep, CALLOUT_MPSAFE);
 	callout_init(&tp->t_timers->tt_2msl, CALLOUT_MPSAFE);
 	callout_init(&tp->t_timers->tt_delack, CALLOUT_MPSAFE);
+	callout_init(&tp->t_timers->tt_nvp, CALLOUT_MPSAFE);
+	callout_init(&tp->t_timers->tt_pack, CALLOUT_MPSAFE);
 
 	if (V_tcp_do_rfc1323)
 		tp->t_flags = (TF_REQ_SCALE|TF_REQ_TSTMP);
@@ -802,6 +804,10 @@ tcp_newtcpcb(struct inpcb *inp)
 	tp->snd_cwnd = TCP_MAXWIN << TCP_MAX_WINSHIFT;
 	tp->snd_ssthresh = TCP_MAXWIN << TCP_MAX_WINSHIFT;
 	tp->t_rcvtime = ticks;
+	if (V_tcp_do_ncwv) {
+		tp->lossflightsize = 0;
+		tp->pipeack = TCP_MAXWIN << TCP_MAX_WINSHIFT;
+	}
 	/*
 	 * IPv4 TTL initialization is necessary for an IPv6 socket as well,
 	 * because the socket may be bound to an IPv6 wildcard address,
@@ -931,6 +937,8 @@ tcp_discardcb(struct tcpcb *tp)
 	callout_stop(&tp->t_timers->tt_keep);
 	callout_stop(&tp->t_timers->tt_2msl);
 	callout_stop(&tp->t_timers->tt_delack);
+	callout_stop(&tp->t_timers->tt_nvp);
+	callout_stop(&tp->t_timers->tt_pack);
 
 	/*
 	 * If we got enough samples through the srtt filter,
diff --git a/sys/netinet/tcp_timer.c b/sys/netinet/tcp_timer.c
index 7c27397..a4f211f 100644
--- a/sys/netinet/tcp_timer.c
+++ b/sys/netinet/tcp_timer.c
@@ -646,6 +646,54 @@ out:
 }
 
 void
+tcp_timer_nvp(void * xtp)
+{
+	struct tcpcb *tp = xtp;
+	struct inpcb *inp;
+	CURVNET_SET(tp->t_vnet);
+	INP_INFO_WLOCK(&V_tcbinfo);
+	inp = tp->t_inpcb;
+	if (inp == NULL) {
+		tcp_timer_race++;
+		INP_INFO_WUNLOCK(&V_tcbinfo);
+		CURVNET_RESTORE();
+		return;
+	}
+	INP_WLOCK(inp);
+	callout_deactivate(&tp->t_timers->tt_nvp);
+	tp->snd_cwnd = max(tp->snd_cwnd / 2 , tp->IW );
+	tp->snd_ssthresh =  max(tp->snd_ssthresh, 3 * tp->snd_cwnd /4);
+	if (tp != NULL)
+		INP_WUNLOCK(inp);
+	INP_INFO_WUNLOCK(&V_tcbinfo);
+	CURVNET_RESTORE();
+}
+
+void
+tcp_timer_pack(void * xtp)
+{
+	struct tcpcb *tp = xtp;
+	struct inpcb *inp;
+	CURVNET_SET(tp->t_vnet);
+	INP_INFO_WLOCK(&V_tcbinfo);
+	inp = tp->t_inpcb;
+	if (inp == NULL) {
+		tcp_timer_race++;
+		INP_INFO_WUNLOCK(&V_tcbinfo);
+		CURVNET_RESTORE();
+		return;
+	}
+	INP_WLOCK(inp);
+	callout_deactivate(&tp->t_timers->tt_pack);
+	tp->snd_cwnd = max(tp->snd_cwnd / 2, tp->IW );
+	tp->snd_ssthresh =  max(tp->snd_ssthresh, 3 * tp->snd_cwnd /4);
+	if (tp != NULL)
+		INP_WUNLOCK(inp);
+	INP_INFO_WUNLOCK(&V_tcbinfo);
+	CURVNET_RESTORE();
+}
+
+void
 tcp_timer_activate(struct tcpcb *tp, int timer_type, u_int delta)
 {
 	struct callout *t_callout;
@@ -679,6 +727,14 @@ tcp_timer_activate(struct tcpcb *tp, int timer_type, u_int delta)
 			t_callout = &tp->t_timers->tt_2msl;
 			f_callout = tcp_timer_2msl;
 			break;
+		case TT_NVP:
+			t_callout = &tp->t_timers->tt_nvp;
+			f_callout = tcp_timer_nvp;
+			break;
+		case TT_PACK:
+			t_callout = &tp->t_timers->tt_pack;
+			f_callout = tcp_timer_pack;
+			break;
 		default:
 			panic("bad timer_type");
 		}
@@ -710,6 +766,12 @@ tcp_timer_active(struct tcpcb *tp, int timer_type)
 		case TT_2MSL:
 			t_callout = &tp->t_timers->tt_2msl;
 			break;
+		case TT_NVP:
+			t_callout = &tp->t_timers->tt_nvp;
+			break;
+		case TT_PACK:
+			t_callout = &tp->t_timers->tt_pack;
+			break;
 		default:
 			panic("bad timer_type");
 		}
@@ -738,5 +800,9 @@ tcp_timer_to_xtimer(struct tcpcb *tp, struct tcp_timer *timer,
 		xtimer->tt_keep = (timer->tt_keep.c_time - now) / SBT_1MS;
 	if (callout_active(&timer->tt_2msl))
 		xtimer->tt_2msl = (timer->tt_2msl.c_time - now) / SBT_1MS;
+	if (callout_active(&timer->tt_nvp))
+		xtimer->tt_nvp = ticks_to_msecs(timer->tt_nvp.c_time - ticks);
+	if (callout_active(&timer->tt_pack))
+		xtimer->tt_pack = ticks_to_msecs(timer->tt_pack.c_time - ticks);
 	xtimer->t_rcvtime = ticks_to_msecs(ticks - tp->t_rcvtime);
 }
diff --git a/sys/netinet/tcp_timer.h b/sys/netinet/tcp_timer.h
index 3115fb3..7d038e8 100644
--- a/sys/netinet/tcp_timer.h
+++ b/sys/netinet/tcp_timer.h
@@ -146,12 +146,16 @@ struct tcp_timer {
 	struct	callout tt_keep;	/* keepalive */
 	struct	callout tt_2msl;	/* 2*msl TIME_WAIT timer */
 	struct	callout tt_delack;	/* delayed ACK timer */
+	struct 	callout tt_nvp;     /* non validated timer period */
+	struct  callout tt_pack; /* timer for pipeack measurement */
 };
 #define TT_DELACK	0x01
 #define TT_REXMT	0x02
 #define TT_PERSIST	0x04
 #define TT_KEEP		0x08
 #define TT_2MSL		0x10
+#define TT_NVP		0x20
+#define TT_PACK		0x40
 
 #define	TP_KEEPINIT(tp)	((tp)->t_keepinit ? (tp)->t_keepinit : tcp_keepinit)
 #define	TP_KEEPIDLE(tp)	((tp)->t_keepidle ? (tp)->t_keepidle : tcp_keepidle)
@@ -183,6 +187,8 @@ void	tcp_timer_keep(void *xtp);
 void	tcp_timer_persist(void *xtp);
 void	tcp_timer_rexmt(void *xtp);
 void	tcp_timer_delack(void *xtp);
+void	tcp_timer_nvp(void *xtp);
+void	tcp_timer_pack(void *xtp);
 void	tcp_timer_to_xtimer(struct tcpcb *tp, struct tcp_timer *timer,
 	struct xtcp_timer *xtimer);
 
diff --git a/sys/netinet/tcp_var.h b/sys/netinet/tcp_var.h
index aaaa4a4..c8c148f 100644
--- a/sys/netinet/tcp_var.h
+++ b/sys/netinet/tcp_var.h
@@ -137,7 +137,7 @@ struct tcpcb {
 					 * for slow start exponential to
 					 * linear switch
 					 */
-	u_long	snd_spare2;		/* unused */
+	u_long	IW;		/* initial cong window */
 	tcp_seq	snd_recover;		/* for use in NewReno Fast Recovery */
 
 	u_int	t_maxopd;		/* mss plus options */
@@ -147,8 +147,9 @@ struct tcpcb {
 	u_int	t_rtttime;		/* RTT measurement start time */
 	tcp_seq	t_rtseq;		/* sequence number being timed */
 
-	u_int	t_bw_spare1;		/* unused */
-	tcp_seq	t_bw_spare2;		/* unused */
+	u_long	lossflightsize;	/* flightsize at the beggining of current recovery event */
+	u_long	pipeack;		/* amount of data acked per RTT */
+	tcp_seq	max_ack_prev;	/* caching of previous value of snd_max when rtt was measured */
 
 	int	t_rxtcur;		/* current retransmit value (ticks) */
 	u_int	t_maxseg;		/* maximum segment size */
@@ -247,6 +248,11 @@ struct tcpcb {
 #define	TF_ECN_SND_ECE	0x10000000	/* ECN ECE in queue */
 #define	TF_CONGRECOVERY	0x20000000	/* congestion recovery mode */
 #define	TF_WASCRECOVERY	0x40000000	/* was in congestion recovery */
+#define TF_RLIMPHASE 0x80000000     /* ncwv phase */
+
+#define	IN_VALPHASE(t_flags)	(t_flags & TF_RLIMPHASE)
+#define	ENTER_VALPHASE(t_flags)	t_flags |= TF_RLIMPHASE
+#define	EXIT_VALPHASE(t_flags)	t_flags &= ~TF_RLIMPHASE
 
 #define	IN_FASTRECOVERY(t_flags)	(t_flags & TF_FASTRECOVERY)
 #define	ENTER_FASTRECOVERY(t_flags)	t_flags |= TF_FASTRECOVERY
@@ -561,6 +567,8 @@ struct xtcp_timer {
 	int tt_keep;	/* keepalive */
 	int tt_2msl;	/* 2*msl TIME_WAIT timer */
 	int tt_delack;	/* delayed ACK timer */
+	int tt_nvp;		/* non-Validation period timer */
+	int tt_pack;	/* pipeack sample timer */
 	int t_rcvtime;	/* Time since last packet received */
 };
 struct	xtcpcb {
@@ -627,8 +635,10 @@ VNET_DECLARE(int, tcp_abc_l_var);
 #define	V_tcp_abc_l_var		VNET(tcp_abc_l_var)
 
 VNET_DECLARE(int, tcp_do_sack);			/* SACK enabled/disabled */
+VNET_DECLARE(int, tcp_do_ncwv);			/* New-CWV enabled/disabled */
 VNET_DECLARE(int, tcp_sc_rst_sock_fail);	/* RST on sock alloc failure */
 #define	V_tcp_do_sack		VNET(tcp_do_sack)
+#define	V_tcp_do_ncwv		VNET(tcp_do_ncwv)
 #define	V_tcp_sc_rst_sock_fail	VNET(tcp_sc_rst_sock_fail)
 
 VNET_DECLARE(int, tcp_do_ecn);			/* TCP ECN enabled/disabled */
